from zlib import compressobj, Z_DEFAULT_COMPRESSION, DEFLATED, MAX_WBITS
from base64 import b64encode

import os
import sys
import sqlite3

from xml.etree import cElementTree as ET
from tempfile import gettempdir
from shutil import copyfile
from datetime import datetime


class StandaloneExecution:
    def __init__(self, application_path, businessperiod):
        self.application_path = application_path
        self.businessperiod = businessperiod
        self.orders_to_proccess = []
        self.orders_not_found = []
        self.orders_not_processed = []
        self.server_directory = os.path.join("data", "server")
        self.update_query_boh = """DELETE FROM EVENTSSENT WHERE PERIOD >= {}""".format(self.businessperiod)
        self.eventtype = "('PAID', 'VOID_ORDER', 'ORDER_REPROCESS')"
        self.auditlogger_query = \
            """SELECT ORDERID, EVTTYPE FROM AUDITLOG WHERE EVTTYPE IN {} 
and BUSINESSPERIOD = {}""".format(self.eventtype, self.businessperiod)
        self.update_query_auditlogger = """UPDATE AUDITLOG SET EVTDATA = '{}' WHERE ORDERID = {} and EVTTYPE = '{}'"""

        self.is_backup = False
        self.auditlog_file = self.file_per_businessperiod()
        self.orders_to_validate = self.orders_on_auditlogger()
        self.boh_file = self.validate_boh_file()


    def make_backup(self):
        if not self.is_backup:
            print("Creating backup files ...\n")
            date_backupfile = datetime.now().strftime('%Y%m%d_%H%M%S')
            self.backup_auditlog = 'auditlog_' + date_backupfile
            self.backup_boh = 'bohpump_' + date_backupfile

            copyfile(self.auditlog_file, gettempdir() + os.path.join(os.sep, self.backup_auditlog))
            copyfile(self.boh_file, gettempdir() + os.path.join(os.sep, self.backup_boh))

            self.is_backup = True


    def validate_boh_file(self):
        boh_directory = os.path.join("bohpump")
        boh_full_path = os.path.join(self.server_directory, boh_directory)
        boh_file_name = "bohpump.ctrl"
        boh_file = os.path.join(self.application_path, boh_full_path, boh_file_name)
        if os.path.isfile(boh_file):
            return boh_file

        raise Exception("No file found: {}".format(boh_file))


    def file_per_businessperiod(self):
        templatelog = "mwapp_audit_{}.log".format(self.businessperiod)
        auditlog_directory = os.path.join("auditlog")
        auditlog_fullpath = os.path.join(self.server_directory, auditlog_directory)
        auditlog_file = os.path.join(self.application_path, auditlog_fullpath, templatelog)
        if os.path.isfile(auditlog_file):
            return auditlog_file

        raise Exception("No file found: {}".format(auditlog_file))


    def orders_on_auditlogger(self):
        print("Verifying orders...")
        orders_to_process = []
        try:
            connection, cursor = self.connection_database(self.auditlog_file)
            for orderid, state in cursor.execute(self.auditlogger_query):
                orders_to_process.append([orderid, state])
        except Exception as e:
            print("Unable to execute query on {}\n {}".format(self.auditlog_file, e))
        finally:
            cursor.close()

        if orders_to_process:
            return orders_to_process

        print("No orders found on: \n{} \nfor businessperiod: {}".format(self.auditlog_file, self.businessperiod))
        sys.exit(1)


    def update_auditlog(self, evt_data, orderid, evt_type):
        update = self.update_query_auditlogger.format(evt_data, orderid, evt_type)
        try:
            connection = sqlite3.connect(self.auditlog_file)
            cursor = connection.cursor()
            cursor.execute(update)
            connection.commit()
            cursor.close()
        except Exception as ex:
            raise ("Unable to update auditlog_file: {}".format(ex))


    def encode_data(self, evt_data):
        encoded = self.compress_data(evt_data)
        encode_data = b64encode(str(encoded))

        return encode_data


    def compress_data(self, evt_data):
        compress_obj = compressobj(Z_DEFAULT_COMPRESSION, DEFLATED, MAX_WBITS + 16)

        return compress_obj.compress(evt_data) + compress_obj.flush()


    def update_bohpump_database(self):
        connection, cursor = self.connection_database(self.boh_file)
        try:
            cursor.execute(self.update_query_boh)
            connection.commit()
            cursor.close()
        except Exception as ex:
            raise Exception("Unable to execute query: \n {}\n{}".format(self.boh_file, ex))


    def connection_database(self, database):
        connection = sqlite3.connect(database)
        cursor = connection.cursor()

        return connection, cursor


    def parse_orderpicture(self, orderpicture):
        order_element = ET.fromstring(orderpicture)
        parsed_orderpicture = ET.tostring(order_element.find('./'))

        return parsed_orderpicture


    def validate_totals(self, orderpicture, orderid):
        order_element = ET.fromstring(orderpicture)

        stateid = order_element.get("stateId")
        total_gross = float(order_element.get("totalGross"))
        total_tender = float(order_element.get("totalTender"))
        discount_amount = float(order_element.get("orderDiscountAmount"))
        change = float(order_element.get("change"))

        totals = total_tender - (total_gross + discount_amount + change)
        if totals != 0.0 and stateid == "5":
            self.orders_not_processed.append(orderid)
            return False

        return True


    def finalization(self):
        backup = "Backup Files saved on: \n{}\n{}\n"
        orders_updated = "Orders updated on mwapp_audit file:\n{}\n"
        restart = "Restart bohpump to send updated orders"
        not_found = "Orders not found on orders databases:\n{}\n"
        not_processed = "Orders not processed with inconsistent orderpicture:\n{}"
        done = "Process done.\n"

        if self.is_backup:
            self.update_bohpump_database()
            print(backup.format(os.path.join(gettempdir(), self.backup_auditlog),
                                os.path.join(gettempdir(), self.backup_boh)))
            print(orders_updated.format(','.join(str(o) for o in self.orders_to_proccess)))

            print(restart)

        print(not_found.format(','.join(str(o) for o in self.orders_not_found)) if self.orders_not_found else '')

        print(not_processed.format(','.join(str(o) for o in self.orders_not_processed)) if self.orders_not_processed else '')

        print(done)


def open_mwpos_connection(info_env, orderid):
    mbcontext = None
    bin_path = os.path.join(info_env.application_path, 'bin')
    common_path = os.path.join("src", "tablemgr", "src", "common")
    os.environ["BINPATH"] = os.path.join(info_env.application_path, bin_path)
    os.environ["HVPORT"] = "14000"
    os.environ["HVIP"] = "127.0.0.1"
    os.environ["HVCOMPPORT"] = "35689"
    os.environ["HVPID"] = "-1"
    sys.path.insert(0, os.path.join(info_env.application_path, common_path))

    try:
        os.chdir(os.environ["BINPATH"])

        from msgbus import MBEasyContext, FM_PARAM, TK_SLCTRL_OMGR_ORDERPICT, TK_SYS_ACK

        mbcontext = MBEasyContext("orderpicture_auditlogger")

        pos_id = "1"
        template_orderid = '\00{}'.format(orderid)
        msg = mbcontext.MB_EasySendMessage(
            "ORDERMGR{}".format(pos_id),
            TK_SLCTRL_OMGR_ORDERPICT,
            format=FM_PARAM,
            data=template_orderid
        )

        if msg.token == TK_SYS_ACK:
            parsed_data = msg.data.split("\0")
            order_picture = parsed_data[2]

        if order_picture:
            return info_env.parse_orderpicture(order_picture)

    except BaseException:
        raise Exception("Please execute with python on application directory")
    finally:
        if mbcontext:
            mbcontext.MB_EasyFinalize()


def application_and_file_path(dev_path=None):
    if dev_path:
        return dev_path

    platform = os.sys.platform
    linux_directory = "/home/administrador/edeployPOS"
    windows_directory = "C:\\edeployPOS"
    if platform == "linux2" and os.path.exists(linux_directory):
        return linux_directory
    elif platform == 'win32' and os.path.exists(windows_directory):
        return windows_directory


if __name__ == "__main__":
    businessperiod = raw_input("Inform a valid businessperiod: ")
    application_directory = \
        raw_input("Press enter to run on PROD Or inform a DEV Path: EX: C:\Dev\edeploy-pos-sctructure\n")

    if not application_directory:
        application_directory = application_and_file_path()

    executor = StandaloneExecution(application_directory, businessperiod)

    for order in executor.orders_to_validate:
        orderid, evt_type = order[0], order[1]
        orderpicture = open_mwpos_connection(executor, orderid)
        if orderpicture and executor.validate_totals(orderpicture, orderid):
            executor.make_backup()
            executor.orders_to_proccess.append(orderid)
            evt_data = executor.encode_data(orderpicture)
            executor.update_auditlog(evt_data, orderid, evt_type)
        elif orderpicture is None:
            executor.orders_not_found.append(orderid)

    executor.finalization()



class LogEntry(object):
    def __init__(
            self,
            business_period,
            event_type,
            pos_id=-1,
            created_at_gmt=None,
            session_id=None,
            event_data='',
            order_id=None
    ):
        # type: (str, str, int, str, str, str, str) -> None
        self.pos_id = pos_id
        self.business_period = business_period
        self.event_type = event_type
        self.created_at_gmt = created_at_gmt
        self.session_id = session_id
        self.event_data = event_data
        self.order_id = order_id



from posot import OrderTaker
from typing import Optional

from open_loyalty import get_document_from_loyalty_or_customer_response
from utilfunctions import (
    fill_additional_info,
    get_custom_property_value,
)


def get_loyalty_document(
    pos_id,                 # type: str
    pos_ot,                 # type: OrderTaker
    ask_document_pinpad,    # type: bool
):
    # type: (...) -> Optional[str]

    loyalty_customer_info = get_custom_property_value(
        pos_id=pos_id,
        custom_property='LOYALTY_CUSTOMER_INFO',
    )

    if loyalty_customer_info is not None:
        document = get_custom_property_value(
            pos_id=pos_id,
            custom_property='CUSTOMER_DOC',
        )
        return document or ""

    loyalty_document = get_document_from_loyalty_or_customer_response(
        pos_id=pos_id,
        pos_ot=pos_ot,
        ask_document_pinpad=ask_document_pinpad,
    )
    if loyalty_document is None:
        return None

    document = loyalty_document
    fill_additional_info(
        pos_ot=pos_ot,
        properties="CPF",
        value=document,
    )

    return document



from msgbus import MBEasyContext
from old_helper import BaseRepository
from typing import List


class UserRepository(BaseRepository):
    def __init__(
            self,
            mb_context  # type: MBEasyContext
    ):

        super(UserRepository, self).__init__(mb_context)

    def get_own_delivery_men(
            self
    ):
        # type: (...) -> List

        def inner_func(conn):

            couriers = []
            query = "SELECT UserId, LongName FROM Users WHERE Level = -1"
            cursor = conn.select(query)
            if cursor.rows() > 0:
                for row in cursor:
                    couriers.append({'userId': row.get_entry("UserId"), 'name': row.get_entry("LongName")})

            return couriers

        return self.execute_with_connection(inner_func)



import sys

if sys.platform == "win32":
    from tzlocal.win32 import (
        get_localzone,
        get_localzone_name,
        reload_localzone,
    )  # pragma: no cover
else:
    from tzlocal.unix import get_localzone, get_localzone_name, reload_localzone


__all__ = ["get_localzone", "get_localzone_name", "reload_localzone"]



from abc import ABCMeta, abstractmethod

from production.model import ProductionOrder


class PublishScheduler(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def schedule_publish(self, order, wait_time):
        # type: (ProductionOrder, int) -> None
        raise NotImplementedError()



import logging

from .dto import LockBenefitRequest
from application.repository.api import APIRepository

logger = logging.getLogger(__name__)


class LockBenefitInteractor(object):
    def __init__(
        self,            # type: LockBenefitInteractor
        api_repository,  # type: APIRepository
    ):
        # type: (...) -> None

        self.api_repository = api_repository

    def lock_benefit(
        self,       # type: LockBenefitInteractor
        request,    # type: LockBenefitRequest
    ):
        # type: (...) -> str

        logger.info("Locking benefit: {}; PosId: {}".format(request.benefit.get("id"), request.pos_id))
        lock_id = self.api_repository.lock_benefit(
            benefit=request.benefit,
            pos_id=request.pos_id,
            payment_info=request.payment_info,
        )
        message = "Benefit {}; PosId: {} locked with lock id: {}"
        logger.info(message.format(request.benefit.get("id"), request.pos_id, lock_id))
        return lock_id



# isin.py - functions for handling ISIN numbers
#
# Copyright (C) 2015-2017 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ISIN (International Securities Identification Number).

The ISIN is a 12-character alpha-numerical code specified in ISO 6166 used to
identify exchange listed securities such as bonds, commercial paper, stocks
and warrants. The number is formed of a two-letter country code, a nine
character national security identifier and a single check digit.

This module does not currently separately validate the embedded national
security identifier part (e.g. when it is a CUSIP).

More information:

* https://en.wikipedia.org/wiki/International_Securities_Identification_Number

>>> validate('US0378331005')
'US0378331005'
>>> validate('US0378331003')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> from_natid('gb', 'BYXJL75')
'GB00BYXJL758'
"""

from stdnum.exceptions import *
from stdnum.util import clean


# all valid ISO 3166-1 alpha-2 country codes
_iso_3116_1_country_codes = [
    'AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AN', 'AO', 'AQ', 'AR', 'AS',
    'AT', 'AU', 'AW', 'AX', 'AZ', 'BA', 'BB', 'BD', 'BE', 'BF', 'BG', 'BH',
    'BI', 'BJ', 'BL', 'BM', 'BN', 'BO', 'BQ', 'BR', 'BS', 'BT', 'BV', 'BW',
    'BY', 'BZ', 'CA', 'CC', 'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM',
    'CN', 'CO', 'CR', 'CS', 'CU', 'CV', 'CW', 'CX', 'CY', 'CZ', 'DE', 'DJ',
    'DK', 'DM', 'DO', 'DZ', 'EC', 'EE', 'EG', 'EH', 'ER', 'ES', 'ET', 'FI',
    'FJ', 'FK', 'FM', 'FO', 'FR', 'GA', 'GB', 'GD', 'GE', 'GF', 'GG', 'GH',
    'GI', 'GL', 'GM', 'GN', 'GP', 'GQ', 'GR', 'GS', 'GT', 'GU', 'GW', 'GY',
    'HK', 'HM', 'HN', 'HR', 'HT', 'HU', 'ID', 'IE', 'IL', 'IM', 'IN', 'IO',
    'IQ', 'IR', 'IS', 'IT', 'JE', 'JM', 'JO', 'JP', 'KE', 'KG', 'KH', 'KI',
    'KM', 'KN', 'KP', 'KR', 'KW', 'KY', 'KZ', 'LA', 'LB', 'LC', 'LI', 'LK',
    'LR', 'LS', 'LT', 'LU', 'LV', 'LY', 'MA', 'MC', 'MD', 'ME', 'MF', 'MG',
    'MH', 'MK', 'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU',
    'MV', 'MW', 'MX', 'MY', 'MZ', 'NA', 'NC', 'NE', 'NF', 'NG', 'NI', 'NL',
    'NO', 'NP', 'NR', 'NU', 'NZ', 'OM', 'PA', 'PE', 'PF', 'PG', 'PH', 'PK',
    'PL', 'PM', 'PN', 'PR', 'PS', 'PT', 'PW', 'PY', 'QA', 'RE', 'RO', 'RS',
    'RU', 'RW', 'SA', 'SB', 'SC', 'SD', 'SE', 'SG', 'SH', 'SI', 'SJ', 'SK',
    'SL', 'SM', 'SN', 'SO', 'SR', 'SS', 'ST', 'SV', 'SX', 'SY', 'SZ', 'TC',
    'TD', 'TF', 'TG', 'TH', 'TJ', 'TK', 'TL', 'TM', 'TN', 'TO', 'TR', 'TT',
    'TV', 'TW', 'TZ', 'UA', 'UG', 'UM', 'US', 'UY', 'UZ', 'VA', 'VC', 'VE',
    'VG', 'VI', 'VN', 'VU', 'WF', 'WS', 'YE', 'YT', 'ZA', 'ZM', 'ZW']

# These special codes are allowed for ISIN
_country_codes = set(_iso_3116_1_country_codes + [
    'EU',  # European Union
    'QS',  # internally used by Euroclear France
    'QS',  # temporarily assigned in Germany
    'QT',  # internally used in Switzerland
    'XA',  # CUSIP Global Services substitute agencies
    'XB',  # NSD Russia substitute agencies
    'XC',  # WM Datenservice Germany substitute agencies
    'XD',  # SIX Telekurs substitute agencies
    'XF',  # internally assigned, not unique numbers
    'XK',  # temporary country code for Kosovo
    'XS',  # international securities
])

# the letters allowed in an ISIN
_alphabet = '0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ'


def compact(number):
    """Convert the number to the minimal representation. This strips the
    number of any valid separators and removes surrounding whitespace."""
    return clean(number, ' ').strip().upper()


def calc_check_digit(number):
    """Calculate the check digits for the number."""
    # convert to numeric first, then double some, then sum individual digits
    number = ''.join(str(_alphabet.index(n)) for n in number)
    number = ''.join(
        str((2, 1)[i % 2] * int(n)) for i, n in enumerate(reversed(number)))
    return str((10 - sum(int(n) for n in number)) % 10)


def validate(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    number = compact(number)
    if not all(x in _alphabet for x in number):
        raise InvalidFormat()
    if len(number) != 12:
        raise InvalidLength()
    if number[:2] not in _country_codes:
        raise InvalidComponent()
    if calc_check_digit(number[:-1]) != number[-1]:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number provided is valid. This checks the length and
    check digit."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def from_natid(country_code, number):
    """Generate an ISIN from a national security identifier."""
    number = country_code.upper() + compact(number).zfill(9)
    return number + calc_check_digit(number)



# Copyright 2015 gRPC authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interfaces related to streams of values or objects."""

import abc


class Consumer(abc.ABC):
    """Interface for consumers of finite streams of values or objects."""

    @abc.abstractmethod
    def consume(self, value):
        """Accepts a value.

        Args:
          value: Any value accepted by this Consumer.
        """
        raise NotImplementedError()

    @abc.abstractmethod
    def terminate(self):
        """Indicates to this Consumer that no more values will be supplied."""
        raise NotImplementedError()

    @abc.abstractmethod
    def consume_and_terminate(self, value):
        """Supplies a value and signals that no more values will be supplied.

        Args:
          value: Any value accepted by this Consumer.
        """
        raise NotImplementedError()



# npwp.py - functions for handling Indonesian NPWP numbers
# coding: utf-8
#
# Copyright (C) 2020 Leandro Regueiro
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""NPWP (Nomor Pokok Wajib Pajak, Indonesian VAT Number).

The Nomor Pokok Wajib Pajak (NPWP) is assigned to organisations and
individuals (families) by the Indonesian Tax Office after registration by the
tax payers.

The number consists of 15 digits of which the first 2 denote the type of
entity, 6 digits to identify the tax payer, a check digit over the first 8
digits followed by 3 digits to identify the local tax office and 3 digits for
branch code.

More information:

* https://www.oecd.org/tax/automatic-exchange/crs-implementation-and-assistance/tax-identification-numbers/Indonesia-TIN.pdf
* https://metacpan.org/pod/Business::ID::NPWP
* https://wiki.scn.sap.com/wiki/display/CRM/Indonesia

>>> validate('01.312.166.0-091.000')
'013121660091000'
>>> validate('016090524017000')
'016090524017000'
>>> validate('123456789')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> format('013000666091000')
'01.300.066.6-091.000'
"""  # noqa: E501

from stdnum import luhn
from stdnum.exceptions import *
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators and removes
    surrounding whitespace.
    """
    return clean(number, ' -.').strip()


def validate(number):
    """Check if the number is a valid Indonesia NPWP number."""
    number = compact(number)
    if len(number) != 15:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    luhn.validate(number[:9])
    return number


def is_valid(number):
    """Check if the number is a valid Indonesia NPWP number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    number = compact(number)
    return '%s.%s.%s.%s-%s.%s' % (
        number[:2], number[2:5], number[5:8], number[8], number[9:12], number[12:])



from typing import Any

from ._BaseMessageProcessor import BaseUpdaterProcessor
from application.interactor import LoaderUpdaterInteractor
from application.util import settings


class LoaderUpdaterProcessor(BaseUpdaterProcessor):
    def __init__(
        self,           # type: LoaderUpdaterProcessor
        interactor,     # type: LoaderUpdaterInteractor
    ):
        # type: (...) -> None
        super(LoaderUpdaterProcessor, self).__init__()

        self.interactor = interactor

    def call_business(
        self,   # type: LoaderUpdaterProcessor
        _,      # type: str
    ):
        # type: (...) -> Any

        settings.clear_status()
        return self.interactor.update_loaders()



# ice.py - functions for handling Morocco ICE numbers
# coding: utf-8
#
# Copyright (C) 2022 Leandro Regueiro
# Copyright (C) 2022 Arthur de Jong
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA
# 02110-1301 USA

"""ICE (Identifiant Commun de l’Entreprise, التعريف الموحد للمقاولة, Morocco tax number).

The ICE is a number that identifies the company and its branches in a unique
and uniform way by all Moroccan administrations. It comes in addition to the
other legal identifiers, notably the "identifiant fiscal" (IF), the "numéro de
registre de commerce" (RC) and the CNSS number. The ICE does not replace these
identifiers, which remain mandatory.

The ICE is intended to ease communication among Moroccan administration
branches, therefore simplifying procedures, increasing reliability and speed,
and therefore reducing costs.

The ICE applies to legal entities and their branches, as well as to natural
persons.

The ICE consists of 15 characters, where the first 9 represent the enterprise,
the following 4 represent its establishments, and the last 2 are control
characters.

More information:

* https://www.ice.gov.ma/
* https://www.ice.gov.ma/ICE/Depliant_ICE.pdf
* https://www.ice.gov.ma/ICE/Guide_ICE.pdf

>>> validate('001561191000066')
'001561191000066'
>>> validate('00 21 36 09 30 00 040')
'002136093000040'
>>> validate('12345')
Traceback (most recent call last):
    ...
InvalidLength: ...
>>> validate('001561191000065')
Traceback (most recent call last):
    ...
InvalidChecksum: ...
>>> format('00 21 36 09 30 00 040')
'002136093000040'
"""

from stdnum.exceptions import *
from stdnum.iso7064 import mod_97_10
from stdnum.util import clean, isdigits


def compact(number):
    """Convert the number to the minimal representation.

    This strips the number of any valid separators, removes surrounding
    whitespace.
    """
    return clean(number, ' ')


def validate(number):
    """Check if the number is a valid Morocco ICE number.

    This checks the length and formatting.
    """
    number = compact(number)
    if len(number) != 15:
        raise InvalidLength()
    if not isdigits(number):
        raise InvalidFormat()
    if mod_97_10.checksum(number) != 0:
        raise InvalidChecksum()
    return number


def is_valid(number):
    """Check if the number is a valid Morocco ICE number."""
    try:
        return bool(validate(number))
    except ValidationError:
        return False


def format(number):
    """Reformat the number to the standard presentation format."""
    return compact(number).zfill(15)



# Copyright The OpenTelemetry Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# FIXME: Use importlib.metadata when support for 3.11 is dropped if the rest of
# the supported versions at that time have the same API.
from importlib_metadata import (  # type: ignore
    EntryPoint,
    EntryPoints,
    entry_points,
    version,
)

# The importlib-metadata library has introduced breaking changes before to its
# API, this module is kept just to act as a layer between the
# importlib-metadata library and our project if in any case it is necessary to
# do so.

__all__ = ["entry_points", "version", "EntryPoint", "EntryPoints"]



import json

from typing import (
    Dict,
)

from .token import TK_SLCTRL_PMGR_GET_TENDERS

from messagebus import (
    MessageBus,
    Message,
    DataType,
    DefaultToken,
)


class ProductService(object):

    def __init__(
        self,               # type: ProductService
        message_bus,        # type: MessageBus
    ):
        # type: (...) -> None

        self.message_bus = message_bus
        self.component_name = "ORDERMGR"

    def get_tenders_list(
        self,       # type: ProductService
    ):
        # type: (...) -> Dict

        msg = self._get_tenders()
        json_data = json.loads(msg.data)
        return _byteify(json_data)

    def get_tenders_dict(
        self,       # type: ProductService
    ):
        # type: (...) -> Dict

        msg = self._get_tenders()
        tenders = json.loads(msg.data)
        tenders_dict = {}
        for tender in tenders:
            tenders_dict[tender["id"]] = tender

        return _byteify(tenders_dict)

    def _get_tenders(
        self,       # type: ProductService
    ):
        # type: (...) -> Message

        msg = self.message_bus.send_message(
            component=self.component_name,
            message=Message(
                token=TK_SLCTRL_PMGR_GET_TENDERS,
                data_type=DataType.param,
            )
        )

        if msg.token != DefaultToken.TK_SYS_ACK:
            raise RuntimeError("Could not retrieve tenders")

        return msg


def _byteify(input):
    if isinstance(input, dict):
        return {_byteify(key): _byteify(value) for key, value in input.iteritems()}

    if isinstance(input, list):
        return [_byteify(element) for element in input]

    if isinstance(input, unicode):
        return input.encode('utf-8')

    return input



from json import JSONEncoder
from typing import (
    Any,
    Dict,
)

from ._Order import Order
from ._OrderPicture import OrderPicture
from ._TablePicture import TablePicture
from ._PaymentPicture import PaymentPicture


class OrderJsonEncoder(JSONEncoder):
    def default(self, o):
        instances_dictionary = {
            Order: self._get_order_json,
            OrderPicture: self._get_order_picture_json,
            TablePicture: self._get_table_picture_json,
            PaymentPicture: self._get_payment_picture_json,
        }

        if type(o) in instances_dictionary:
            return instances_dictionary[type(o)](o)

        return super(OrderJsonEncoder, self).default(o)

    @staticmethod
    def _get_order_json(
        o,  # type: Order
    ):
        # type: (...) -> Dict[str, Any]

        return {
            "orderPicture": o.order_picture,
            "tablePicture": o.table_picture,
            "paymentPicture": o.payment_picture,
        }

    @staticmethod
    def _get_order_picture_json(
        o,  # type: OrderPicture
    ):
        # type: (...) -> Dict[str, Any]

        return {
            "storeCode": o.store_code,
            "orderInfo": o.order_info,
        }

    @staticmethod
    def _get_table_picture_json(
        o,  # type: TablePicture
    ):
        # type: (...) -> Dict[str, Any]

        return {
            "storeCode": o.store_code,
            "tableInfo": o.table_info,
        }

    @staticmethod
    def _get_payment_picture_json(
        o,    # type: PaymentPicture
    ):
        # type: (...) -> Dict[str, Any]

        return {
            "storeCode": o.store_code,
            "paymentInfo": o.payment_info,
        }



import base64
import json
from datetime import datetime
from typing import (
    Dict,
    List,
    Optional,
    Any,
)

from xml.etree import cElementTree as eTree

from ._Type import JSON
from ._util import get_tender_detail


class OrderFormatter:

    def __init__(self):
        return

    def get_order_picture_dict(
        self,                   # type: OrderFormatter
        order_picture,          # type: eTree.Element
    ):
        # type: (...) -> JSON

        session_id = self._get_order_picture_property(
            order_picture=order_picture,
            order_property="sessionId",
        )
        sale_type = int(order_picture.get("saleType", 0))

        order_picture_dict = {
            "orderId": str(order_picture.get("orderId")),
            "businessPeriod": self._get_business_period(order_picture=order_picture),
            "change": float(order_picture.get("change", 0)),
            "createdAt": order_picture.get("createdAt"),
            "createdAtGMT": order_picture.get("createdAtGMT") + "Z",
            "discountAmount": float(order_picture.get("discountAmount", 0)),
            "discountsApplied": order_picture.get("discountsApplied"),
            "dueAmount": float(order_picture.get("dueAmount", 0)),
            "instance": int(order_picture.get("instance", 0)),
            "lastModifiedLine": int(order_picture.get("lastModifiedLine", 0)),
            "lineCounter": int(order_picture.get("lineCounter", 0)),
            "orderDiscountAmount": float(order_picture.get("orderDiscountAmount", 0)),
            "originatorId": str(int(order_picture.get("originatorId")[-2:])),
            "podType": order_picture.get("podType"),
            "posId": self._get_pos_code(
                pos_id=order_picture.get("posId"),
                instance=order_picture.get("instance"),
                session_id=session_id,
            ),
            "priceBasis": order_picture.get("priceBasis"),
            "priceList": order_picture.get("priceList"),
            "saleType": sale_type if sale_type >= 0 else 0,
            "saleTypeDescr": order_picture.get("saleTypeDescr"),
            "sessionId": session_id,
            "stateId": int(order_picture.get("stateId")),
            "state": order_picture.get("stateDescr"),
            "subType": order_picture.get("subType"),
            "taxTotal": float(order_picture.get("taxTotal", 0)),
            "tip": float(order_picture.get("tip", 0)),
            "totalAfterDiscount": float(order_picture.get("totalAfterDiscount", 0)),
            "totalAmount": float(order_picture.get("totalAmount", 0)),
            "totalGross": float(order_picture.get("totalGross", 0)),
            "totalTender": float(order_picture.get("totalTender", 0)),
            "type": order_picture.get("type"),
            "typeId": int(order_picture.get("typeId")),
            "userId": self._get_pos_user_id(session_id=session_id),
            "CustomOrderProperties": self._get_custom_properties(
                order_picture=order_picture,
            ),
            "SaleLine": self._get_sale_lines(order_picture=order_picture),
            "StateHistory": self._get_state_history(order_picture=order_picture),
            "TenderHistory": self._get_order_tenders(order_picture=order_picture),
            "xml": self._get_fiscal_xml(
                order_picture=order_picture,
                fiscal_type="FISCAL_XML",
            ),
            "xmlCancel": self._get_fiscal_xml(
                order_picture=order_picture,
                fiscal_type="CANCELED_FISCAL_XML",
            ),
        }

        return order_picture_dict

    @staticmethod
    def _get_fiscal_xml(
        order_picture,  # type: eTree.Element
        fiscal_type,    # type: str
    ):
        # type: (...) -> Optional[str]

        fiscal_xml = None
        prop = order_picture.find(".//OrderProperty[@key='{}']".format(fiscal_type))
        if prop is not None:
            fiscal_xml = prop.get("value")

        if fiscal_xml is None:
            return None

        fiscal_xml = fiscal_xml.replace(" ", "")
        decoded_fiscal_xml = base64.b64decode(fiscal_xml)
        return base64.b64encode(decoded_fiscal_xml)

    @staticmethod
    def _get_order_picture_property(
        order_picture,      # type: eTree.Element
        order_property,     # type: str
    ):
        # type: (...) -> Optional[str]

        return order_picture.get("{}".format(order_property))

    @staticmethod
    def _get_business_period(
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> str

        order_business_period = order_picture.get("businessPeriod")
        return datetime.strptime(order_business_period, "%Y%m%d").strftime("%Y-%m-%d")

    @staticmethod
    def _get_pos_user_id(
        session_id,     # type: str
    ):
        # type: (...) -> str

        if session_id == "":
            return '-1'

        for session_data in session_id.split(","):
            key, value = session_data.split("=")
            if key == "user":
                return value

    @staticmethod
    def _get_pos_code(
        pos_id,         # type: str
        instance,       # type: str
        session_id,     # type: str
    ):
        # type: (...) -> str

        if pos_id is None and session_id != "":
            for session_data in session_id.split(","):
                key, value = session_data.split("=")
                if key == "pos":
                    pos_id = value
                    break

        if pos_id is None:
            pos_id = instance

        return pos_id

    @staticmethod
    def _get_custom_properties(
        order_picture,      # type: eTree.Element
    ):
        # type: (...) -> Dict

        payload_custom_order_properties = {}

        custom_properties = order_picture.find("CustomOrderProperties") or []
        for custom_property in custom_properties:
            if custom_property.get("key").upper() not in ["FISCAL_XML", "CANCELED_FISCAL_XML"]:
                custom_property_key = custom_property.get("value").encode("utf-8")
                payload_custom_order_properties[custom_property.get("key")] = custom_property_key

        return payload_custom_order_properties

    @staticmethod
    def _get_order_tenders(
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> List

        payload_order_tenders = []
        for tender_line in order_picture.findall("TenderHistory/Tender"):
            tender_payload = {
                "tenderId": tender_line.get("tenderId"),
                "tenderAmount": float(tender_line.get("tenderAmount", 0)),
                "tenderDescr": tender_line.get("TenderDescr"),
                "tenderType": tender_line.get("tenderType"),
                "changeAmount": float(tender_line.get("change", 0)),
                "tenderDetail": get_tender_detail(tender_line=tender_line),
                "tenderTip": float(tender_line.get("tip")) if tender_line.get("tip") else None
            }
            payload_order_tenders.append(tender_payload)

        return payload_order_tenders

    def _get_sale_lines(
        self,
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> List[Dict[str, Any[str, int]]]

        payload_order_sale_lines = []
        for sale_line in order_picture.findall("SaleLine"):
            sale_line_payload = {
                "addedUnitPrice": float(sale_line.get("addedUnitPrice", 0)),
                "decQty": float(sale_line.get("decQty", 0)),
                "incQty": float(sale_line.get("incQty", 0)),
                "itemId": sale_line.get("itemId"),
                "itemPrice": float(sale_line.get("itemPrice", 0)),
                "itemType": sale_line.get("itemType"),
                "lastUpdated": sale_line.get("lastUpdated"),
                "level": int(sale_line.get("level")),
                "lineNumber": int(sale_line.get("lineNumber")),
                "measureUnit": sale_line.get("measureUnit"),
                "multipliedQty": float(sale_line.get("multipliedQty") or sale_line.get("qty") or 0),
                "partCode": sale_line.get("partCode"),
                "priceKey": sale_line.get("priceKey"),
                "productName": sale_line.get("productName").encode("utf-8"),
                "productPriority": int(sale_line.get("productPriority")),
                "qty": float(sale_line.get("qty", 0)),
                "unitPrice": float(sale_line.get("unitPrice", 0)),
                "defaultQty": float(sale_line.get("defaultQty")) if sale_line.get("defaultQty") else None,
                "addedQty": float(sale_line.get("addedQty")) if sale_line.get("addedQty") else None,
                "subQty": float(sale_line.get("subQty")) if sale_line.get("subQty") else None,
                "chosenQty": float(sale_line.get("chosenQty")) if sale_line.get("chosenQty") else None,
                "discountsApplied": sale_line.get("discountsApplied"),
                "itemDiscount": float(sale_line.get("itemDiscount", 0)),
                "custom-properties": self._get_sale_line_custom_properties(sale_line=sale_line),
                "taxItem": self._get_sale_line_tax_items(sale_line=sale_line),
                "comment": self._get_sale_line_comments(sale_line=sale_line),
                "subUnitPrice": float(sale_line.get("subUnitPrice", 0)),
                "tags": self._get_sale_line_tags(sale_line=sale_line),
            }
            payload_order_sale_lines.append(sale_line_payload)

        return payload_order_sale_lines

    @staticmethod
    def _get_state_history(
        order_picture,  # type: eTree.Element
    ):
        # type: (...) -> JSON

        list_state_history = []
        for state in order_picture.find("StateHistory").findall("State"):
            list_state_history.append({
                "state": state.get("state"),
                "stateId": state.get("stateId"),
                "timestamp": state.get("timestamp"),
                "timestampGMT": state.get("timestampGMT"),
            })
        return list_state_history

    @staticmethod
    def _get_sale_line_comments(
        sale_line,  # type: eTree.Element
    ):
        # type: (...) -> List

        comments = []
        for comment in sale_line.findall("Comment"):
            comments.append(comment.get("comment").encode("utf-8"))

        return comments

    @staticmethod
    def _get_sale_line_custom_properties(
        sale_line,  # type: eTree.Element
    ):
        # type: (...) -> JSON

        custom_properties = {}
        try:
            custom_properties = json.loads(sale_line.get("customProperties", {}))
        except TypeError:
            pass
        return custom_properties

    @staticmethod
    def _get_sale_line_tax_items(
        sale_line,  # type: eTree.Element
    ):
        # type: (...) -> JSON

        tax_items = []
        for order_tax_item in sale_line.findall("TaxItem"):
            tax_item = {
                "baseAmountAD": float(order_tax_item.get("baseAmountAD", 0)),
                "baseAmountBD": float(order_tax_item.get("baseAmountBD", 0)),
                "taxAmountAD": float(order_tax_item.get("taxAmountAD", 0)),
                "taxAmountBD": float(order_tax_item.get("taxAmountBD", 0)),
                "taxIndex": order_tax_item.get("taxIndex"),
                "taxRate": float(order_tax_item.get("taxRate", 0)),
                "taxRuleId": int(order_tax_item.get("taxRuleId"))
            }
            tax_items.append(tax_item)

        return tax_items

    @staticmethod
    def _get_sale_line_tags(
        sale_line,  # type: eTree.Element
    ):
        # type: (...) -> List

        item_tags = []
        try:
            json_array_tags = json.loads(sale_line.get("jsonArrayTags", "[]"))
            for temp_tag in json_array_tags:
                item_tags.append(temp_tag)
        except TypeError:
            pass

        return item_tags



import logging
from concurrent.futures import ThreadPoolExecutor
from threading import Thread

from messagebus import (
    MessageBus,
    DefaultToken,
)
from opentelemetry import trace
from opentelemetry.trace import StatusCode
from timeutil import Clock
from . import EventProcessorListener
from ._EventPublishDetail import EventPublishDetail

tracer = trace.get_tracer(__name__)
logger = logging.getLogger(__name__)


class Publisher(object):
    def __init__(
        self,
        message_bus: MessageBus,
        event_processor_listener: EventProcessorListener,
        clock: Clock,
        max_workers: int = 10,
    ):
        self.message_bus = message_bus
        self.event_processor_listener = event_processor_listener
        self.clock = clock

        self.pool = ThreadPoolExecutor(
            max_workers=max_workers,
            thread_name_prefix="Publisher",
        )

    def publish_event(
        self,
        publish_detail: EventPublishDetail,
    ) -> None:

        call_callback = self._build_callback(publish_detail=publish_detail)

        future = self.pool.submit(self._publish_event, publish_detail)
        future.add_done_callback(fn=call_callback)

    def finish(self):
        thread = Thread(
            target=self.pool.shutdown,
            name="FinishThread",
        )
        thread.daemon = True
        logger.info("Stopping Publisher")
        thread.start()

        thread.join(timeout=2)

    def _publish_event(self, publish_detail: EventPublishDetail):
        with tracer.start_span("AsyncPublisher") as span:
            span.set_attribute("definitionId", str(publish_detail.id))

            try:
                if publish_detail.target is not None:
                    reply = self.message_bus.send_message(
                        component=publish_detail.target,
                        message=publish_detail.event.convert_to_message()
                    )
                else:
                    reply = self.message_bus.publish_event(event=publish_detail.event)

                if publish_detail.has_async_target():
                    publish_detail.calculate_next_retry_at(now=self.clock.utc_now())
                    span.set_status(StatusCode.OK)
                else:
                    if reply.token == DefaultToken.TK_SYS_ACK:
                        publish_detail.acked_at = self.clock.utc_now()
                        span.set_status(StatusCode.OK)
                    else:
                        publish_detail.calculate_next_retry_at(now=self.clock.utc_now())
                        span.set_status(StatusCode.ERROR)
            except (Exception,):
                publish_detail.calculate_next_retry_at(now=self.clock.utc_now())

    def _build_callback(
        self,
        publish_detail: EventPublishDetail,
    ):
        def finished_processing(_):
            self.event_processor_listener.event_processed(publish_detail=publish_detail)

        return finished_processing
